# Transforming Growth Rate Parameters
# Code by Blair Fix

#######################################################################################
# This code is designed to transform the Laplace growth rate parameters from a semi-annual 
# time interval to the associated parameters of an annual time interval (which is a more useful).
# For every given location parameter, the code calculates the relation between 
# the standard deviation of a semi-annual growth rates and the standard deviation of annual growth rates.
# The relation between annual/semi-annual scale parameters is is equivalent to relation between the two standard deviations.
# The location parameter of an annual compounding period is the square of the location parameter of a semi-annual compounding period.


library(doSNOW)
library(foreach)
library(snow)
library(poweRlaw)
library(tcltk)
library(VGAM)

# Clusters
##################################################
cl <- makeCluster(4, type="SOCK")
registerDoSNOW(cl)
clusterSetupRNG (cl, type = "RNGstream")


# Lower Boundary
#######################################
# log.seq = a function to generate log-spaced sequences
vector.length = 960
log.seq = function(x,y,z) { exp(seq(log(x), log(y), length.out = z))}
loc.1 = 1.01
loc.2 = 0.75
location.vector = log.seq(loc.1, loc.2, vector.length)
z = 1:vector.length/vector.length
scale.vector =  - 10^(-0.83)*(z-1)^10 + 0.19

# Upper Boundary
#######################################
#   vector.length = 960
#   loc.1 = 1.01
#   loc.2 = 0.9
#   location.vector = log.seq(loc.1, loc.2, vector.length)
z = 1:vector.length/vector.length
#   scale.vector = -10^(-0.67)*(1-z)^1.6 + 0.35

# Calculation variables
#############################
# l = the number of location parameters to sample
l = 40
# n = the number of scale parameters to sample for each location parameter
n = 100
# sd.ratio = variable for the ratio of annual / semi-annual standard deviations
sd.ratio = rep(NA,n)
# b.vec = a vector of scale parameters 
b.vec = seq(0.05,0.4, length.out = n)
# space =  a spacing vector of length l used to select l location parameters
space = round(seq(1,vector.length, length.out = l ))
# u.vec = a vector of location parameters
u.vec = location.vector[space]


# Progress Bar
####################################
startTime <- proc.time()
pb <- txtProgressBar(max=l, style=3)
progress <- function(n) setTxtProgressBar(pb, n)
opts <- list(progress=progress)

# The code has two nested loops:
# A big loop over location parameters (implemented in parallel) and a small loop over scale parameters
# For each location and scale parameter, two independent random Laplace distributions (x and y) are generated and then multiplied together.
# This gives growth rates over one year.
# Because the relation is stochastic, we run a polynomial regression on the relation between 
# annual/semi-annual standard deviation ratios vs. semi-annual scale parameter (for fixed location parameter)

# big loop over location vector
f =  foreach (h = 1:l, .options.snow=opts,  .combine=rbind, .packages=c("VGAM") ) %dopar%  {

    # u = selected location parameter
    u = u.vec[h]
      
      for(i in 1:n){
      # b = selected scale parameter (semi-annual)  
      b = b.vec[i]
      
      # x = growth rate distribution 1
      x = rlaplace(10^5, u, b)
      # y = grouwth rate distribution 2
      y = rlaplace(10^5, u, b)
      # z = resulting annual growth rate distribution (product of x and y)
      z = x*y
      
      # sd.ratio = the standard deviation ratios of z to x
      sd.ratio[i] = sd(z)/sd(x)
      }
    
    # r = a polynomial regression between sd ratios and scale parameters  
    r = lm(sd.ratio ~ poly(b.vec, degree = 5))
    # print regression coefficients
    as.numeric(coef(r))

}
##############################

stopCluster(cl)
endTime <- proc.time()
time = endTime-startTime
close(pb)

# Apply the regression to predict annual scale parameter
##########################################################
# exp = polynomial exponents 
exp = 0:5
# Initialize sd.ratio vector
sd.ratio = rep(NA, l)
# b.vec = scale parameters to be transformed
b.vec = scale.vector[space]

# loop over scale parameters and find predicted sd ratio
for(i in 1:l){
sd.ratio[i] = sum(f[i,]*b.vec[i]^exp)
}

# b.vec = vector of transformed scale parameters
b.vec = b.vec*sd.ratio
# u.vec = vector of transformed location parameters
u.vec = u.vec^2

# plots
#################################
plot(u.vec, b.vec, type = "l", xlim = c(0.9, 1.1))




